package pluck;

import java.io.PrintWriter;
import java.math.BigInteger;
import java.net.SocketException;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Map.Entry;

import snmp.SNMPBERCodec;
import snmp.SNMPBadValueException;
import snmp.SNMPCounter32;
import snmp.SNMPGetException;
import snmp.SNMPInteger;
import snmp.SNMPObject;
import snmp.SNMPObjectIdentifier;
import snmp.SNMPOctetString;
import snmp.SNMPPDU;
import snmp.SNMPRequestListener;
import snmp.SNMPSequence;
import snmp.SNMPSetException;
import snmp.SNMPTimeTicks;
import snmp.SNMPVariablePair;
import snmp.SNMPv1AgentInterface;

public class SNMPAgent implements SNMPRequestListener{
	private class SOIDComparator implements Comparator<String>{
		@Override
		public int compare(String o1, String o2) {
			if(o1.equals(o2)) return 0;
			//System.err.println(o1.substring(o1.lastIndexOf(".")+1)+" "+o2.substring(o2.lastIndexOf(".")+1));			
			//return o1.compareTo(o2);
			if(o1.substring(0,o1.lastIndexOf(".")).equals(o2.substring(0,o2.lastIndexOf("."))))		return Integer.signum(
				Integer.parseInt(o1.substring(o1.lastIndexOf(".")+1)) - Integer.parseInt(o2.substring(o2.lastIndexOf(".")+1)));
			else return o1.compareTo(o2);
		}
	}
	private static final String ReleaseCauseBase = "1.3.6.1.4.1.65065.2.2";
	private static final String CallsNumBase ="1.3.6.1.4.1.65065.1.2.2";
	private static final String RefsNumBase ="1.3.6.1.4.1.65065.1.2.3";
	static final BigInteger incVal=new BigInteger("1");
    private long startTime;
    private SortedMap<String,SNMPObject> oMap;
    private String communityName = "public";
    public SNMPAgent(int port,String community, Properties props) throws SocketException{
	communityName=community;
	startTime=Calendar.getInstance().getTimeInMillis();
   	 oMap = Collections.synchronizedSortedMap(new TreeMap<String,SNMPObject>(new SOIDComparator()));
	 oMap.put("1.3.6.1.2.1.1.1.0",new SNMPOctetString("Pluck v 0.2"));
	 oMap.put("1.3.6.1.2.1.1.3.0", new SNMPTimeTicks((Calendar.getInstance().getTimeInMillis()-startTime)/10));    	 
	 oMap.put("1.3.6.1.2.1.1.4.0", new SNMPOctetString(props.getProperty("sysContact", "Admin@")));    	 
	 oMap.put("1.3.6.1.2.1.1.5.0",new SNMPOctetString(props.getProperty("sysName", "Server")));
	 oMap.put("1.3.6.1.2.1.1.6.0",new SNMPOctetString(props.getProperty("sysLocation", "Earth")));
	 oMap.put(CallsNumBase+".1",new SNMPInteger(0));
	 oMap.put(RefsNumBase+".1",new SNMPInteger(0));
	 

	 initPortIdxs(oMap);
	 //oMap.put("1.3.6.1.4.1.65065.2.2.1.1",new SNMPCounter32(120));    	 
	 //oMap.put("1.3.6.1.4.1.65065.2.2.1.1151",new SNMPCounter32(55));
	 		SNMPv1AgentInterface agentInterface = new SNMPv1AgentInterface(0, port, new PrintWriter(System.err));
		agentInterface.addRequestListener(this);
     	agentInterface.setReceiveBufferSize(5120);
     	agentInterface.startReceiving();
    }
	private void initPortIdxs(SortedMap<String, SNMPObject> oMap2) {
		oMap.put("1.3.6.1.4.1.65065.2.1.1.1",new SNMPInteger(1));
 		oMap.put("1.3.6.1.4.1.65065.1.2.1.1",new SNMPOctetString("Total"));
 		initPortCauses(1);

		for(int p=1;p<9;p++){
		 for(int m=1;m<5;m++){
		  oMap.put("1.3.6.1.4.1.65065.2.1.1."+getIdx(p, m),new SNMPInteger(getIdx(p, m)));
		  oMap.put("1.3.6.1.4.1.65065.1.2.1."+getIdx(p, m),new SNMPOctetString("Port "+m+"/"+p+" ("+((m-1)*8+p)+")"));
		  initPortCauses(getIdx(p, m));
		  }
		 for(int m=9;m<13;m++){
		  oMap.put("1.3.6.1.4.1.65065.2.1.1."+getIdx(p, m),new SNMPInteger(getIdx(p, m)));
		  oMap.put("1.3.6.1.4.1.65065.1.2.1."+getIdx(p, m),new SNMPOctetString("Port "+m+"/"+p+" ("+((m-5)*8+p)+")"));
		  initPortCauses(getIdx(p, m));
		  }
		}
/*
		for(int p=1;p<33;p++) oMap.put("1.3.6.1.4.1.65065.2.1.2."+p,new SNMPInteger(getIdx(1+(p-1)%8, 1+(p-1)/8)));
		for(int p=33;p<65;p++) oMap.put("1.3.6.1.4.1.65065.2.1.2."+p,new SNMPInteger(getIdx(1+(p-1)%8, 1+(p-1)/8+4)));
*/
	}
	private void initPortCauses(int idx) {
		for(int c=1;c<128;c++)	oMap.put(ReleaseCauseBase+"."+c+"."+idx,new SNMPCounter32(0));
	}
	private int getIdx(int p, int m) {
		return (10+m)*100+p+50;
	}
	@Override
	public SNMPSequence processGetNextRequest(SNMPPDU pdu,
			String communityName) throws SNMPGetException {
		//printDump(pdu);
		SNMPSequence responseList = new SNMPSequence();
		
		if(!authorized(pdu)) return responseList; 
        
        SNMPSequence varBindList = pdu.getVarBindList();
        
        for (int i = 0; i < varBindList.size(); i++)
        {
            SNMPSequence variablePair = (SNMPSequence)varBindList.getSNMPObjectAt(i);
            SNMPObjectIdentifier suppliedOID = (SNMPObjectIdentifier)variablePair.getSNMPObjectAt(0);
        		//System.err.println("rqNext: "+suppliedOID);
            	SortedMap<String,SNMPObject> sMap=oMap.tailMap(suppliedOID.toString());
            	if(sMap==null) continue;
            	String key;
            	SNMPObject value;
            	if(sMap.firstKey().equals(suppliedOID.toString())){
            	//System.err.println(" 3"+sMap.firstKey());
            		Iterator<Entry<String, SNMPObject>> sMapIter=sMap.entrySet().iterator();
            		sMapIter.next();
            		if(!sMapIter.hasNext()) continue;
            		key=sMapIter.next().getKey();
            		value=sMap.get(key);
           		//System.err.println(" 4"+key);
            	}else{
            	 key=sMap.firstKey();
            	 value=sMap.get(key);
            	}
                    try
                    {
                        // create SNMPVariablePair for the next OID and its value 
                        SNMPObjectIdentifier nextOID = new SNMPObjectIdentifier(key);
                        SNMPVariablePair innerPair = new SNMPVariablePair(nextOID, value);
                        
                        // now create a pair containing the supplied OID and the variable pair containing the following
                        // OID and its value; this allows the ANMPv1AgentInterface to know which of the supplied OIDs
                        // the new OID corresponds to (follows).
                        SNMPVariablePair outerPair = new SNMPVariablePair(suppliedOID, innerPair);
                        
                        // add the "compound" SNMPVariablePair to the response list
                        responseList.addSNMPObject(outerPair);
                    }
                    catch (SNMPBadValueException e)
                    {
                        // won't happen...
                    }
            	
            	
/*            	
            	for (Map.Entry<String,SNMPObject> pair:sMap.entrySet()){
            		System.err.println(pair.getKey()+" "+pair.getValue());
            	}
*/
        }
        writeMessage("\n");
        // return the created list of variable pairs
        return responseList;
	}
	@Override
	public SNMPSequence processRequest(SNMPPDU pdu, String communityName)
			throws SNMPGetException, SNMPSetException {
		//printDump(pdu);
		SNMPSequence responseList = new SNMPSequence();
		
		if(!authorized(pdu)) return responseList; 
        
        SNMPSequence varBindList = pdu.getVarBindList();
        
        for (int i = 0; i < varBindList.size(); i++)
        {
            SNMPSequence variablePair = (SNMPSequence)varBindList.getSNMPObjectAt(i);
            SNMPObjectIdentifier snmpOID = (SNMPObjectIdentifier)variablePair.getSNMPObjectAt(0);
            
            SNMPObject o=oMap.get(snmpOID.toString());
            if(o==null) continue;
            try
            {
            	SNMPVariablePair newPair = new SNMPVariablePair(new SNMPObjectIdentifier(snmpOID.toString()), o);
                //SNMPVariablePair newPair = new SNMPVariablePair(snmpOID, new SNMPOctetString("Boo"));
                responseList.addSNMPObject(newPair);
            }
            catch (SNMPBadValueException e)
            {
                // won't happen...
            }
        }
        writeMessage("\n");
        return responseList;
	}
	
    private boolean authorized(SNMPPDU pdu) {
        if (!communityName.equals(this.communityName)) 	return false;
        byte pduType = pdu.getPDUType();
        if (pduType != SNMPBERCodec.SNMPGETREQUEST&&pduType != SNMPBERCodec.SNMPGETNEXTREQUEST) return false;
		return true;
	}
	private void printDump(SNMPPDU pdu) {
			        writeMessage("Got pdu:\n");
        writeMessage("  community name:     " + communityName + "\n");
        writeMessage("  request ID:         " + pdu.getRequestID() + "\n");
        writeMessage("  pdu type:           ");
        byte pduType = pdu.getPDUType();
        switch (pduType)
        {
            case SNMPBERCodec.SNMPGETREQUEST:
            {
                writeMessage("SNMPGETREQUEST\n");
                break;
            }
            case SNMPBERCodec.SNMPGETNEXTREQUEST:
            {
                writeMessage("SNMPGETNEXTREQUEST\n");
                break;
            }
            case SNMPBERCodec.SNMPSETREQUEST:
            {
                writeMessage("SNMPSETREQUEST\n");
                break;
            }
            case SNMPBERCodec.SNMPGETRESPONSE:
            {
                writeMessage("SNMPGETRESPONSE\n");
                break;
            }
            case SNMPBERCodec.SNMPTRAP:
            {
                writeMessage("SNMPTRAP\n");
                break;
            }
            default:
            {
                writeMessage("unknown\n");
                break;
            }
        }
        SNMPSequence varBindList = pdu.getVarBindList();
        for (int i = 0; i < varBindList.size(); i++)
        {
            SNMPSequence variablePair = (SNMPSequence)varBindList.getSNMPObjectAt(i);
            SNMPObjectIdentifier snmpOID = (SNMPObjectIdentifier)variablePair.getSNMPObjectAt(0);
            SNMPObject snmpValue = (SNMPObject)variablePair.getSNMPObjectAt(1);
            writeMessage("       OID:           " + snmpOID + "\n");
            writeMessage("       value:         " + snmpValue + "\n");
		}
	}
	private void writeMessage(String message)
    {
		System.err.print(message);
    }
	public void incReleaseCauseCount(int cause) {
		SNMPObject o=oMap.get(ReleaseCauseBase+"."+cause+".1");
        if(o==null) oMap.put(ReleaseCauseBase+"."+cause+".1",new SNMPCounter32(1));
        else SNMPCounter32.class.cast(o).add(incVal);
	}
	public void incReleaseCauseCount(int cause,int loc) {
		SNMPObject o=oMap.get(ReleaseCauseBase+"."+cause+"."+(1000+loc));
        if(o==null) oMap.put(ReleaseCauseBase+"."+cause+"."+(1000+loc),new SNMPCounter32(1));
        else SNMPCounter32.class.cast(o).add(incVal);
	}

	public void setCheckups(int calls,int refs) {
		try {
			SNMPObject o=oMap.get(CallsNumBase+".1");
			SNMPInteger.class.cast(o).setValue(calls);
			o=oMap.get(RefsNumBase+".1");
			SNMPInteger.class.cast(o).setValue(refs);
		} catch (SNMPBadValueException e) {
			e.printStackTrace();
		}
		
	}
}
